#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <signal.h>

#include "core_c2bf.h"
#include "network.h"
#include "rcfile.h"
#include "c2.h"

static void c2bf();
static void reset_work_block(WORK_BLOCK *work);
static WORK_UNIT *select_work_unit(WORK_BLOCK *work);
static void update(WORK_UNIT *unit, int idx);
static void interval(int count, char *msg);
	
static void teardown(int sig);
static int global_finish;

int main(int argc, char **argv)
{
	/* set idle priority */
	setpriority(PRIO_PROCESS, getpid(), 20);

	/* set signal handler */
	global_finish = 0;
	if(signal(SIGINT, teardown) == SIG_ERR){
		fprintf(stderr, "failed on signal(SIGINT)\n");
	}
	if(signal(SIGTERM, teardown) == SIG_ERR){
		fprintf(stderr, "failed on signal(SIGTERM)\n");
	}
	c2_init();
	
	c2bf();
	
	return EXIT_SUCCESS;
}

static void teardown(int sig)
{
	global_finish = 1;
}

static void c2bf()
{
	int n;
	int i,j,max;
	
	uint32_t sec;
	uint32_t now;
	
	WORK_BLOCK work;
	WORK_INFO  info;
	WORK_UNIT *unit;
	
	memset(&work, 0, sizeof(WORK_BLOCK));
	memset(&info, 0, sizeof(WORK_INFO));

	if(!load(&work)){
		i = 0;
		while( (n = get_id(&work)) < 0 ){
			interval(i, "[E] failed on get_id()");
			i += 1;
			if(global_finish){
				return;
			}
		}
		if(n == 0){
			fprintf(stdout, "[E] failed on get_id()\n");
			fflush(stdout);
			return;
		}
	}

	save(&work);

	do {
		unit = select_work_unit(&work);
		if( (unit == NULL) && (work.count) ){
			i = 0;
			while( (n = report(&work)) < 0 ){
				interval(i, "[E] failed on report()");
				i += 1;
				if(global_finish){
					return;
				}
			}
			if(n == 0){
				work.count = 0;
				save(&work);
				fprintf(stdout, "[E] failed on report()\n");
				fflush(stdout);
				return;
			}
			fprintf(stdout, "rank: %s\n", work.rank);
			fflush(stdout);
		}
		now = time(NULL);
		if( (unit == NULL) || (now-86400 > work.time) ){
			reset_work_block(&work);
			save(&work);
			i = 0;
			while( (n = get_block(&work)) < 0 ){
				interval(i, "[E] failed on get_block()");
				i += 1;
				if(global_finish){
					return;
				}
			}
			if(n == 0){
				fprintf(stdout, "[E] failed on get_block()\n");
				fflush(stdout);
				return;
			}
			if(n == 2){
				fprintf(stdout, "found key=%014llx\n", work.unit[0].k[0]);
				fflush(stdout);
				return;
			}
			unit = select_work_unit(&work);
		}
		unit->st = 1;
		save(&work);
		info.pa = (uint32_t)((work.plaintext>>32) & 0xffffffff);
		info.pb = (uint32_t)( work.plaintext      & 0xffffffff);
		info.ka = unit->ka;
		info.kb = unit->kb;

		for(i=0;i<8;i++){
			info.ca[i] = (uint32_t)((unit->c[i]>>32) & 0xffffffff);
			info.cb[i] = (uint32_t)( unit->c[i]      & 0xffffffff);
		}

		max = 256 - ((info.kb>>24) & 0xff);
		for(i=0;i<max;i++){
			for(j=0;j<256*256*256;j+=4){
				n = check_c2_4key_mmx(&info);
				if(n){
					update(unit, n);
				}
				info.kb += 4;
				unit->kb += 4;
			}
			save(&work);
			if(global_finish){
				return;
			}
		}

		sec = time(NULL) - now;
		fprintf(stdout, "block: %06x - elapsed %d sec\n", unit->ka, sec*256/max);
		fflush(stdout);
		
		unit->st = 2;
		save(&work);
		
	}while(!global_finish);
}

static void reset_work_block(WORK_BLOCK *work)
{
	int i;

	work->plaintext = 0;
	work->time = 0;
	work->count = 0;
	
	for(i=0;i<8;i++){
		memset(work->unit+i, 0, sizeof(WORK_UNIT));
	}
}

static WORK_UNIT *select_work_unit(WORK_BLOCK *work)
{
	int i;

	for(i=0;i<work->count;i++){
		if(work->unit[i].st != 2){
			return work->unit+i;
		}
	}

	return NULL;
}

static void update(WORK_UNIT *unit, int idx)
{
	int n;
	int i,j;

	for(i=0;i<4;i++){
		n = (idx >> (i*8)) & 0xff;
		j = -1;
		switch(n){
		case 0x01:
			j = 0;
			break;
		case 0x02:
			j = 1;
			break;
		case 0x04:
			j = 2;
			break;
		case 0x08:
			j = 3;
			break;
		case 0x10:
			j = 4;
			break;
		case 0x20:
			j = 5;
			break;
		case 0x40:
			j = 6;
			break;
		case 0x80:
			j = 7;
			break;
		}
		if(j >= 0){
			unit->k[j] = ((uint64_t)unit->ka)<<32;
			unit->k[j] |= unit->kb;
			unit->k[j] |= i;
			unit->ff |= (1<<j);
		}
	}
}

static void interval(int count, char *msg)
{
	int i,n;

	i = count;
	if(i > 8){
		i = 8;
	}
	n = 1<<i;

	fprintf(stdout, "%s - %d, wait %d sec\n", msg, count, n*15);
	fflush(stdout);

	for(i=0;i<n;i++){
		sleep(15);
		if(global_finish){
			break;
		}
	}
}

	
